import os

import pytest
from rara_tools.converters import SierraResponseConverter
from rara_tools.exceptions import SierraResponseConverterException

from tests.const import SIERRA_OUTPUT_DIR
from tests.test_utils import (read_json_file, get_formatted_sierra_response, compare_results)


example_res = {
    "total": 100,
    "start": 50000,
    "entries": [
        {
            "id": 1126963,
            "updatedDate": "2016-02-09T08:42:52Z",
            "createdDate": "2014-05-17T17:22:00Z",
            "deleted": False,
            "suppressed": False,
            "marc": {
                "leader": "00000nz  a2200145n  4500",
                "fields": [
                    {
                        # "tag": "100",
                        "data": {
                            "ind1": "1",
                                    "ind2": " ",
                                    "subfields": [
                                        {
                                            "code": "a",
                                            "data": "Viggor, Signe,"
                                        },
                                        {
                                            "code": "d",
                                            "data": "1975-"
                                        }
                                    ]
                        }
                    },
                ]}}]}


def test_convert_bibs_response():

    data = get_formatted_sierra_response("bibs.json")

    expected = read_json_file(os.path.join(SIERRA_OUTPUT_DIR, "bibs.json"))

    assert compare_results(expected, data)


def test_convert_keywords_response():

    data = get_formatted_sierra_response("keywords.json")

    expected = read_json_file(os.path.join(SIERRA_OUTPUT_DIR, "keywords.json"))

    assert compare_results(expected, data)


def test_convert_authorities_response():

    data = get_formatted_sierra_response("authorities.json")

    expected = read_json_file(os.path.join(
        SIERRA_OUTPUT_DIR, "authorities.json"))

    assert compare_results(expected, data)


def test_converter_handles_marc_in_json_response():
    """ Gracefully handle entries already in MARC-in-JSON format """
    data = get_formatted_sierra_response("bibsmarc.json")

    expected = read_json_file(os.path.join(SIERRA_OUTPUT_DIR, "bibsmarc.json"))

    assert compare_results(expected, data)


def test_convert_with_wrong_format():
    with pytest.raises(SierraResponseConverterException):
        SierraResponseConverter("$")


def test_convert_missing_tag():
    with pytest.raises(SierraResponseConverterException):
        response = example_res.copy()
        response["entries"][0]["marc"]["fields"][0].pop("tag", None)

        converter = SierraResponseConverter(response)
        converter.convert()


def test_no_entries_in_response():
    with pytest.raises(SierraResponseConverterException):
        response = example_res.copy()
        response.pop("entries", [])

        converter = SierraResponseConverter(response)
        converter.convert()



import pytest
import os
from rara_tools.parsers.marc_parsers.ems_parser import EMSMARCParser
from rara_tools.parsers.marc_parsers.person_parser import PersonsMARCParser
from rara_tools.parsers.marc_parsers.organization_parser import OrganizationsMARCParser
from rara_tools.parsers.marc_parsers.location_parser import LocationMARCParser


MARC_ROOT_DIR = os.path.join("tests", "test_data", "marc_records")
EMS_TEST_FILE = os.path.join(MARC_ROOT_DIR, "ems_test_subset.mrc")
PER_TEST_FILE = os.path.join(MARC_ROOT_DIR, "per_test_subset.mrc")
ORG_TEST_FILE = os.path.join(MARC_ROOT_DIR, "org_test_subset.mrc")

def test_ems_parser_without_variations():
    ems_marc_parser = EMSMARCParser(EMS_TEST_FILE, add_variations=False)
    for record in ems_marc_parser.record_generator():
        assert "keyword" in record
        assert "link_variations" not in record

def test_ems_parser_with_variations():
    ems_marc_parser = EMSMARCParser(EMS_TEST_FILE, add_variations=True)
    record_count = 0
    for record in ems_marc_parser.record_generator():
        assert "keyword" in record
        assert "link_variations" in record
        record_count+=1
    assert record_count == 10

def test_loc_parser_with_variations():
    loc_marc_parser = LocationMARCParser(EMS_TEST_FILE, add_variations=True)
    record_count = 0
    for record in loc_marc_parser.record_generator():
        assert "keyword" in record
        assert "link_variations" in record
        record_count+=1
    assert record_count == 5

def test_persons_parser_without_variations():
    per_marc_parser = PersonsMARCParser(PER_TEST_FILE, add_variations=False)
    for record in per_marc_parser.record_generator():
        assert "name" in record
        assert "link_variations" not in record


def test_organizations_parser_without_variations():
    org_marc_parser = OrganizationsMARCParser(ORG_TEST_FILE, add_variations=False)
    for record in org_marc_parser.record_generator():
        assert "name" in record
        assert "link_variations" not in record



from pymarc import (Field, Indicators, Subfield, Record)

from rara_tools.constants import EMPTY_INDICATORS
from rara_tools.normalizers.viaf import VIAFRecord
from rara_tools.normalizers import RecordNormalizer

from typing import List


class BibRecordNormalizer(RecordNormalizer):
    """ Normalize bib records. """

    def __init__(self, linking_results: List[dict] = [], sierra_data: List[dict] = [],
                 ALLOW_EDIT_FIELDS: List[str] = ["667", "925"],
                 REPEATABLE_FIELDS: List[str] = []):
        super().__init__(linking_results, sierra_data)
        self.ALLOW_EDIT_FIELDS = ALLOW_EDIT_FIELDS
        self.REPEATABLE_FIELDS = REPEATABLE_FIELDS

    def _normalize_sierra(self, record: Record) -> Record:
        fields = [
            Field(
                tag="008",
                indicators=EMPTY_INDICATORS,
                data=f"{self.current_timestamp()} | | | aznnnaabn | | | | |"
            ),
            Field(
                tag="046",
                indicators=EMPTY_INDICATORS,
                subfields=[
                    Subfield("k", "Pub date")
                ]
            ),
            Field(
                tag="245",
                indicators=Indicators("1", "0"),
                subfields=[
                    Subfield("a", "Title")
                ]
            ),
        ]

        self._add_fields_to_record(record, fields)

    def _normalize_viaf(self, record: Record, viaf_record: VIAFRecord) -> None:

        if not viaf_record:
            return record

        viaf_id = viaf_record.viaf_id
        fields = [
            Field(
                tag="035",
                indicators=EMPTY_INDICATORS,
                subfields=[
                    Subfield("a", viaf_id)
                ]
            ),
            Field(
                tag="100",
                indicators=EMPTY_INDICATORS,
                subfields=[
                    Subfield("a", "?")
                ]
            )]

        self._add_fields_to_record(record, fields)
        self._add_author(record, viaf_record)

    def _normalize_record(self, record: Record, sierraID: str,
                          viaf_record: VIAFRecord, is_editing_existing_record: bool) -> Record:

        self._normalize_sierra(record)
        self._normalize_viaf(record, viaf_record)

        return record



import regex as re


class Transliterate():
    """
    Transliterate class to transliterate text from Cyrillic to Latin

    Ruleset: http://www.eki.ee/books/ekk09/index.php?id=37&p=2&p1=6
    'Vene keele tähestikust eesti tähestikku' 2005.

    Usage:
        Initialize the class:
            foo = Transliterate()
        Call with a list of sentences you'd like to transliterate:
            foo(['Дженни играет с мяц.', 'Яне говорит на телефоне.', 'Рассекречены планы выпуска дешевого iPhone Читать далее'])
            # This Returns -- ['Дженни играет с мяц.', 'Яне говорит на телефоне.', 'Рассекречены планы выпуска дешевого iPhone Читать далее']
    """

    def __call__(self, sentences):
        return self._transliterate(sentences)

    def _transliterate(self, sentences):
        """The main transliteration processor

        Arguments:
            sentences {list of strings} -- A list of string sentences to tranliterate

        Returns:
            [list of strings] -- A transliterated version of the output sentences
        """
        new_sentences = []
        for sentence in sentences:
            converted_sentence = []
            sentence = sentence.split(' ')
            for word in sentence:
                word = self._check_doublechar_rules(word)
                converted_word = ''
                for char_index, char in enumerate(word):
                    transchar = ''
                    if char in self._translit_table:
                        transchar = self._translit_table[char]
                    elif char in self._chars_with_rules:
                        transchar = self._check_char_rules(
                            word, char, char_index)
                    else:
                        transchar = char

                    converted_word += transchar
                converted_sentence.append(converted_word)
            new_sentences.append(' '.join(converted_sentence))

        return new_sentences

    def _check_doublechar_rules(self, word):
        """Checks for doublechar rules, currently only for the cyr characters ij

        Arguments:
            word {string} -- The word in which to check the rule in

        Returns:
            [string] -- The modified version of the input word relative to the rules, if no rules were found, returns the unchanged input word.
        """
        word_to_return = word
        if len(word) > 3:
            for match in re.finditer("ий", word.lower()):
                if match.span()[1] - 1 == len(word) - 1:
                    new_char = self._check_if_upper_and_return(
                        word[match.span()[0]], 'I')
                    word_to_return = list(word_to_return[:match.span()[0] + 1])
                    word_to_return[-1] = new_char
                    word_to_return = ''.join(word_to_return)

        return word_to_return

    def _check_char_rules(self, word, char, char_index):
        """Checks for specific rules for the given transliteration

        Arguments:
            word {string} -- The word in which the character originated from.
            char {string} -- The character about to be transliterated.
            char_index {int} -- The index of the character in the word.
        """
        if char in 'Ее':
            return self._rules_for_e(word, char, char_index)
        elif char in 'Ёё':
            return self._rules_for_jo(word, char, char_index)
        elif char in 'ИиЙй':
            return self._rules_for_i(word, char, char_index)
        elif char in 'Хх':
            return self._rules_for_h(word, char, char_index)
        elif char in 'Ьь':
            return self._rules_for_soft(word, char, char_index)
        elif char in 'Сс':
            return self._rules_for_s(word, char, char_index)

    def _rules_for_s(self, word, char, char_index):
        """Checks rules for the cyr s character
            Returns: [string] -- Returns 'Ss/ss/S/s' based on the rules
        """
        char_to_return = ''
        if char_index > 0:
            if word[char_index-1] in self._vocals_rus:
                if len(word) == 2:
                    char_to_return = self._check_if_upper_and_return(
                        char, 'Ss')
                # To avoid keyerror use len(word[char_index+1:])
                elif len(word[char_index+1:]) > 0 and word[char_index+1] in self._vocals_rus:
                    char_to_return = self._check_if_upper_and_return(
                        char, 'Ss')
                elif word[-1] == char:
                    char_to_return = self._check_if_upper_and_return(
                        char, 'Ss')

        if char_to_return == '':
            char_to_return = self._chars_with_rules[char]

        return char_to_return

    def _rules_for_soft(self, word, char, char_index):
        """Checks rules for the cyr softening character
            Returns: [string] -- Returns 'J/j//' based on the rules
        """
        if len(word[char_index+1:]) > 0 and word[char_index+1] not in 'еЕёЁюЮяЯ' and word[char_index+1] in self._vocals_rus:
            char_to_return = self._check_if_upper_and_return(char, 'J')
        else:
            return self._chars_with_rules[char]

        return char_to_return

    def _rules_for_e(self, word, char, char_index):
        """Checks rules for the cyr character 'e'
            Returns: [string] -- Returns 'Je/je/E/e' based on the rules
        """
        if char_index == 0:
            char_to_return = self._check_if_upper_and_return(char, 'Je')
        elif word[char_index - 1] in self._vocals_rus or word[char_index - 1] in 'ЬьЪъ':
            char_to_return = self._check_if_upper_and_return(char, 'Je')
        else:
            char_to_return = self._chars_with_rules[char]

        return char_to_return

    def _rules_for_jo(self, word, char, char_index):
        """Checks rules for the cyr character 'jo'
            Returns: [string] -- Returns 'O/o/Jo/jo' based on the rules
        """
        char_to_return = ''
        if char_index != 0:
            if word[char_index-1] in 'ЖжЧчШшЩщ':
                char_to_return = self._check_if_upper_and_return(char, 'O')

        if char_to_return == '':
            char_to_return = self._chars_with_rules[char]

        return char_to_return

    def _rules_for_i(self, word, char, char_index):
        """Checks rules for the cyr character 'i'
        Returns: [string] -- Returns 'J/j/I/i' based on the rules
        """
        char_to_return = ''
        if len(word) > 1:
            if char_index == 0 and word[char_index + 1] in self._vocals_rus:
                char_to_return = self._check_if_upper_and_return(char, 'J')

        if char_to_return == '':
            char_to_return = self._chars_with_rules[char]

        return char_to_return

    def _rules_for_h(self, word, char, char_index):
        """Checks rules for the cyr character 'h'
        Returns: [string] -- Returns 'Hh/hh/H/h' based on the rules
        """
        char_to_return = ''
        if char_index > 0:
            if word[char_index-1] in self._vocals_rus:
                if len(word) == 2:
                    char_to_return = self._check_if_upper_and_return(
                        char, 'Hh')
                # To avoid keyerror use len(word[char_index+1:])
                elif len(word[char_index+1:]) > 0 and word[char_index+1] in self._vocals_rus:
                    char_to_return = self._check_if_upper_and_return(
                        char, 'Hh')
                elif word[-1] == char:
                    char_to_return = self._check_if_upper_and_return(
                        char, 'Hh')

        if char_to_return == '':
            char_to_return = self._chars_with_rules[char]

        return char_to_return

    @staticmethod
    def _check_if_upper_and_return(char_in_index, char):
        """Checks whether the character is supposed to be lower or upper case

        Arguments:
            char_in_index {string} -- [The value of the char index in the word]
            char {string} -- [The character to return, should be first letter uppercase, ex: 'Je']

        Returns:
            [string] -- [Upper or lower case version of the character]
        """
        if char_in_index.isupper():
            return char
        else:
            return char.lower()

    # Ugly constants too big for init method
    _vocals_est = ['i', 'ü', 'u', 'e', 'ö', 'õ', 'o', 'ä', 'a']
    _vocals_rus = ['а', 'э', 'ы', 'у', 'о', 'я', 'е', 'ё', 'ю', 'и']
    _translit_table = {
        'А': 'A', 'а': 'a',
        'Б': 'B', 'б': 'b',
        'В': 'V', 'в': 'v',
        'Г': 'G', 'г': 'g',
        'Д': 'D', 'д': 'd',
        'Ж': 'Ž', 'ж':  'ž',
        'З': 'Z', 'з': 'z',
        'К': 'K', 'к': 'k',
        'Л': 'L', 'л': 'l',
        'М': 'M', 'м': 'm',
        'Н': 'N', 'н': 'n',
        'О': 'O', 'о': 'o',
        'П': 'P', 'п': 'p',
        'Р': 'R', 'р': 'r',
        'Т': 'T', 'т': 't',
        'У': 'U', 'у': 'u',
        'Ф': 'F', 'ф': 'f',
        'Ц': 'Ts', 'ц':  'ts',
        'Ч': 'Tš', 'ч':  'tš',
        'Ш': 'Š', 'ш':  'š',
        'Щ': 'Štš', 'щ': 'štš',
        'Ы': 'õ', 'ы': 'õ',
        'Ъ': "", 'ъ': "",
        'Э': 'E', 'э': 'e',
        'Ю': 'Ju', 'ю': 'ju',
        'Я': 'Ja', 'я': 'ja'}
    _chars_with_rules = {
        'Е': 'E', 'е': 'e',
        'Ё': 'Jo', 'ё': 'jo',
        'И': 'I', 'и': 'i',
        'Й': 'I', 'й': 'i',
        'Х': 'H', 'х':  'h',
        'Ь': "", 'ь': "",
        'С': 'S', 'с': 's',
    }



from typing import NoReturn
from collections.abc import Iterator
from rara_tools.parsers.marc_parsers.base_parser import BaseMARCParser
from rara_tools.parsers.marc_records.ems_record import EMSRecord
from rara_tools.constants.parsers import KeywordType, LOGGER


class EMSMARCParser(BaseMARCParser):
    """ MARC parser for EMS .mrc files.
    """
    def __init__(self,
            marc_file_path: str,
            add_variations: bool = True
        ) -> NoReturn:
        """ Initializes EMSMARCParser object.

        Parameters
        -----------
        marc_file_path: str
            Full path to .mrc file containing EMS data.
        add_variations: bool
            If enabled, constructs an additional variations field, which
            combines the content of multiple fields + adds some generated
            variations. If the output is uploaded into Elastic and used
            via rara-norm-linker, it is necessary to enable this.
        """
        super().__init__(
            marc_file_path=marc_file_path,
            add_variations=add_variations
        )

    def record_generator(self) -> Iterator[EMSRecord]:
        """ Generates EMSRecord objects for topic, genre,
        and time keywords. Location keywords are ignored here
        and handled in a separate parser.
        """
        LOGGER.info(
            f"Generating EMS records (without location keywords) " \
            f"from MARC dump '{self.marc_file_path}'."
        )
        for record in self.marc_record_generator():
            try:
                ems_record = EMSRecord(
                    record=record,
                    add_variations=self.add_variations
                )
                if ems_record.keyword_type == KeywordType.LOC:
                    continue
                else:
                    yield ems_record.full_record
            except Exception as e:
                LOGGER.warning(
                    f"Error while parsing record {record}: {e}"
                )
                continue



import functools
from typing import Any, Callable

from elasticsearch import AuthenticationException
from elasticsearch import ConnectionError as ElasticsearchConnectionError
from elasticsearch import ConnectionTimeout, NotFoundError, RequestError

from .exceptions import ElasticsearchException

ELASTIC_NOT_FOUND_MESSAGE = 'Could not find specified data from Elasticsearch!'
ELASTIC_REQUEST_ERROR_MESSAGE = 'Error executing Elasticsearch query! Bad query?'
ELASTIC_CONNECTION_TIMEOUT_MESSAGE = 'Connection to Elasticsearch took too long, please try again later!'
ELASTIC_AUTHENTICATION_ERROR_MESSAGE = 'Could not authenticate with Elasticsearch!'
ELASTIC_UNKNOWN_ERROR_MESSAGE = 'Unexpected error from Elasticsearch!'
ELASTIC_CONNECTION_ERROR_MESSAGE = 'Could not connect to Elasticsearch, is the location properly configured?'


def _elastic_connection(func: Callable) -> Callable:
	@functools.wraps(func)
	def wrapper(*args: Any, **kwargs: Any) -> Any:
		try:
			return func(*args, **kwargs)
		except NotFoundError as exception:
			raise ElasticsearchException(ELASTIC_NOT_FOUND_MESSAGE) from exception
		except RequestError as exception:
			raise ElasticsearchException(ELASTIC_REQUEST_ERROR_MESSAGE) from exception
		except ConnectionTimeout as exception:
			raise ElasticsearchException(ELASTIC_CONNECTION_TIMEOUT_MESSAGE) from exception
		except AuthenticationException as exception:
			raise ElasticsearchException(ELASTIC_AUTHENTICATION_ERROR_MESSAGE) from exception
		# Important to set the ConnectionError to the bottom of the chain
		# as it's one of the superclasses the other exceptions inherit.
		except ElasticsearchConnectionError as exception:
			if exception.__context__ and 'timed out' in str(exception.__context__):
				# urllib3.exceptions.ConnectTimeoutError can cause an
				# elasticsearch.exceptions.ConnectionError,
				# but we'd like to treat timing out separately
				raise ElasticsearchException(ELASTIC_CONNECTION_TIMEOUT_MESSAGE) from exception
			raise ElasticsearchException(ELASTIC_CONNECTION_ERROR_MESSAGE) from exception
		except Exception as exception:
			raise ElasticsearchException(ELASTIC_UNKNOWN_ERROR_MESSAGE) from exception
	return wrapper



from typing import List, NoReturn
from pymarc.record import Record
from rara_tools.parsers.tools.entity_normalizers import PersonNormalizer
from rara_tools.parsers.marc_records.base_record import BaseRecord
from rara_tools.constants.parsers import PersonMarcIDs, LOGGER
import regex as re
import json
import logging


class PersonRecord(BaseRecord):
    """ Generates a simplified organization JSON record
    from a pymarc MARC record.
    """
    def __init__(self, record: Record, add_variations: bool = False) -> NoReturn:
        """ Initializes PersonRecord object.

        Parameters
        -----------
        record: Record
            pymarc.record.Record object.
        add_variations: bool
            If enabled, constructs an additional variations field, which
            combines the content of multiple fields + adds some generated
            variations. If the output is uploaded into Elastic and used
            via rara-norm-linker, it is necessary to enable this.
        """
        super().__init__(record=record, add_variations=add_variations)

        self.__name_field_id: List[str] = PersonMarcIDs.NAME
        self.__name_variations_field_id: List[str]= PersonMarcIDs.NAME_VARIATIONS
        self.__source_field_id: List[str] = PersonMarcIDs.SOURCE
        self.__description_field_id: List[str] = PersonMarcIDs.DESCRIPTION
        self.__default_year: int | None = None

        self.__name: str = ""
        self.__original_name: dict = {}
        self.__name_specification: str = ""
        self.__life_years: str = ""
        self.__birth_year: int = -1
        self.__death_year: int = -1
        self.__name_variations: List[str] = []
        self.__source: str = ""
        self.__description: str = ""
        self.__full_record: dict = {}
        self.__name_in_cyrillic: bool = None
        self.__variations: List[str] = []
        self.__person_normalizer: PersonNormalizer = PersonNormalizer(self.name)


    def _parse_year(self, year: str) -> int:
        year = year.strip()
        _year = self.__default_year
        if len(year) >= 4:
            if year[:4].isnumeric():
                _year = int(year[:4])
        elif len(year) == 3 and year.isnumeric():
            _year = int(year)
        return _year

    @property
    def original_name(self) -> str:
        if not self.__original_name:
            values = self.get_values(
                marc_ids=self.__name_field_id,
                subfield_id=["a", "b"]
            )
            if values:
                self.__original_name = {
                    "a": self._clean_value(values[0].get("a", "")),
                    "b": self._clean_value(values[0].get("b", ""))
                }
            else:
                pass
        return self.__original_name

    @property
    def name(self) -> str:
        if not self.__name:
            self.__name = self._merge_and_clean(self.original_name, ["a", "b"])
        return self.__name


    @property
    def name_specification(self) -> str:
        if not self.__name_specification:
            values = self.get_values(
                marc_ids=self.__name_field_id,
                subfield_id="c"
            )
            self.__name_specification = self._clean_value(values[0]) if values else ""
        return self.__name_specification

    @property
    def life_years(self) -> str:
        if not self.__life_years:
            values = self.get_values(
                marc_ids = self.__name_field_id,
                subfield_id="d"
            )
            self.__life_years = self._clean_value(values[0]) if values else ""
        return self.__life_years


    @property
    def birth_year(self) -> int:
        if self.__birth_year == -1:
            try:
                birth_year, death_year = self.life_years.split("-")
                self.__birth_year = self._parse_year(birth_year)
                self.__death_year = self._parse_year(death_year)
            except Exception as e:
                LOGGER.error(
                    f"Failed extracting birth and/or death year " \
                    f"from '{self.life_years}' with the following " \
                    f"exception: '{e}'."
                )
        return self.__birth_year


    @property
    def death_year(self) -> int:
        if self.__death_year == -1:
            try:
                birth_year, death_year = self.life_years.split("-")
                self.__birth_year = self._parse_year(birth_year)
                self.__death_year = self._parse_year(death_year)
            except Exception as e:
                LOGGER.error(
                    f"Failed extracting birth and/or death year " \
                    f"from '{self.life_years}' with the following " \
                    f"exception: '{e}'."
                )
        return self.__death_year

    @property
    def name_variations(self) -> List[str]:
        if not self.__name_variations:
            values = self.get_values(
                marc_ids=self.__name_variations_field_id,
                subfield_id=["a", "b"]
            )
            if values:
                raw_variations = [
                    {
                        "a": self._clean_value(value.get("a", "")),
                        "b": self._clean_value(value.get("b", ""))
                    }
                    for value in values
                ]
                self.__name_variations = [
                    self._merge_and_clean(value, ["a", "b"])
                    for value in raw_variations
                ]
            else:
                pass
        return self.__name_variations

    @property
    def source(self) -> str:
        if not self.__source:
            values = self.get_values(
                marc_ids=self.__source_field_id,
                subfield_id="a"
            )
            self.__source = self._clean_value(values[0]) if values else ""
        return self.__source

    @property
    def description(self) -> str:
        if not self.__description:
            values = self.get_values(
                marc_ids=self.__description_field_id,
                subfield_id="i"
            )
            self.__description = self._clean_value(values[0]) if values else ""
        return self.__description

    @property
    def name_in_cyrillic(self) -> bool:
        if self.__name_in_cyrillic == None:
            self.__name_in_cyrillic = PersonNormalizer.has_cyrillic(self.name)
        return self.__name_in_cyrillic

    @property
    def variations(self) -> List[str]:
        if not self.__variations:
            variations_ = self.__person_normalizer.variations
            for name in self.name_variations:
                variations_.extend(PersonNormalizer(name).variations)
            self.__variations = [v.lower() for v in list(set(variations_))]
        return self.__variations

    @property
    def full_record(self) -> dict:
        if not self.__full_record:
            self.__full_record = {
                "name": self.name,
                "life_year": self.life_years,
                "source": self.source,
                "birth_year": self.birth_year,
                "death_year": self.death_year,
                "identifier": self.identifier,
                "identifier_source": self.identifier_source,
                "name_variations": self.name_variations,
                "name_specification": self.name_specification,
                "description": self.description,
                "name_in_cyrillic": self.name_in_cyrillic,
                "full_record_marc": str(self.marc_record),
                "full_record_json": json.dumps(self.marc_json_record)
            }
            if self.add_variations:
                self.__full_record.update(
                    {"link_variations": self.variations}
                )

        return self.__full_record



import pytest
from collections import Counter
from rara_tools.constants.parsers import KeywordType
from rara_tools.parsers.tools.entity_normalizers import PersonNormalizer, KeywordNormalizer


def test_generating_person_variations():
    pn_1 = PersonNormalizer(name="Карл Ристикиви")
    var_1 = pn_1.variations

    pn_2 = PersonNormalizer(name="Ристикиви, Карл")
    var_2 = pn_2.variations

    expected_output = [
        "Karl Ristikivi",
        "Карл Ристикиви",
        "Ристикиви, Карл",
        "Ristikivi, Karl"
    ]
    assert Counter(var_1) == Counter(expected_output)
    assert Counter(var_2) == Counter(expected_output)


def test_generating_keyword_variations():
    kn_1 = KeywordNormalizer(keyword="agendad (religioon)")
    var_1 = kn_1.variations

    expected_output = [
        "agendad (religioon)",
        "agenda",
        "agendad",
        "agenda ( religioon )"
    ]
    assert Counter(var_1) == Counter(expected_output)


def test_generating_loc_keyword_variations():
    # If keyword type != LOC, variations with v -> w
    # replacements should NOT be generated
    kn_1 = KeywordNormalizer(
        keyword="Võrumaal",
        keyword_type=KeywordType.TOPIC
    )
    var_1 = kn_1.variations
    expected_output_1 = [
        "Võrumaal",
        "Võrumaa"
    ]
    assert Counter(var_1) == Counter(expected_output_1)

    # If keyword type == LOC, variations with v -> w
    # replacements should be generated
    kn_2 = KeywordNormalizer(
        keyword="Võrumaal",
        keyword_type=KeywordType.LOC
    )
    var_2 = kn_2.variations
    expected_output_2 = [
        "Võrumaal",
        "Võrumaa",
        "Wõrumaa",
        "Wõrumaal"
    ]
    assert Counter(var_2) == Counter(expected_output_2)



from collections import defaultdict
from typing import List, NoReturn

from rara_tools.utils import lang_to_iso639_2, ratio_to_percentage


UNDEFINED_LANGUAGE_VALUE = "unk"
QUALITY_RATIO_TYPE = "Float"


class ImagePageSchema:
    def __init__(self, image: dict) -> NoReturn:
        self.__image = image
        self.__schema: dict = {}

    @property
    def schema(self) -> dict:
        if not self.__schema:
            self.__schema = {
                "@type": "VisualArtwork",
                "@id": "",
                "value": self.__image.get("label"),
                "description": "",
                "schema:position": self.__image.get("page")
            }
        return self.__schema


class TextPageSchema:
    def __init__(self, page: dict) -> NoReturn:
        self.__page: dict = page
        self.__schema: dict = {}

    @property
    def schema(self) -> dict:
        if not self.__schema:
            self.__schema = {
                "@type": "Text",  # CONSTANT
                "@id": "",  # Will be added in a later stage
                "value": "Textblock",  # CONSTANT
                "content": self.__page.get("text"),
                "schema:position": self.__page.get("start_page")  # start_page ?
            }
        return self.__schema


class PageSchema:
    def __init__(
            self,
            page_texts: List[dict],
            page_images: List[dict],
            page_number: int,
            doc_id: str
    ) -> NoReturn:
        self.__page_texts: List[dict] = page_texts
        self.__page_images: List[dict] = page_images
        self.__page_nr: int = page_number
        self.__page_id: str = ""
        self.__doc_id: str = doc_id
        self.__schema: dict = {}

    def _add_segment_ids(self, segments: List[dict]) -> List[dict]:
        for i, segment in enumerate(segments):
            segment_id = f"{self.page_id}/{i + 1}"
            segment["@id"] = segment_id
        return segments

    @property
    def page_id(self) -> str:
        if not self.__page_id:
            self.__page_id = f"{self.__doc_id}/{self.__page_nr}"
        return self.__page_id

    @property
    def schema(self) -> dict:
        if not self.__schema:
            self.__schema = {
                "@type": "CreativeWork",  # CONSTANT for pages
                "@id": self.page_id,
                "hasPart": []
            }
            text_schemas = [
                TextPageSchema(page).schema
                for page in self.__page_texts
            ]
            image_schemas = [
                ImagePageSchema(image).schema
                for image in self.__page_images
            ]

            page_schemas = text_schemas + image_schemas
            page_schemas_with_ids = self._add_segment_ids(page_schemas)

            self.__schema["hasPart"].extend(page_schemas_with_ids)

        return self.__schema


class DocSchemas:
    def __init__(
            self,
            doc_meta: dict,
            sierra_id: str = "",
            generated_id: str = "",
            permalink: str = "",
            min_language_ratio: float = 0.2,
            convert_ratio: bool = True,
            generated_id_type: str = "CustomID"
    ) -> NoReturn:
        self.__convert_ratio = convert_ratio
        self.__min_language_ratio = min_language_ratio
        self.__sierra_id = sierra_id
        self.__generated_id = generated_id
        self.__permalink = permalink
        self.__generated_id_type = generated_id_type
        self.__doc_meta = doc_meta
        self.__ocr_accuracy_schema: dict = {}
        self.__text_quality_schema: dict = {}
        self.__language_schema: List[dict] = []
        self.__identifier_schema: List[dict] = []
        self.__origin_schema: dict = {}
        self.__origin: str = ""

    @property
    def origin(self) -> str:
        if not self.__origin:
            if self.__doc_meta["ocr_applied"]:
                self.__origin = "Reformatted digital"
            else:
                self.__origin = "Born digital"
        return self.__origin

    @property
    def ocr_accuracy_schema(self) -> dict:
        if not self.__ocr_accuracy_schema:
            ocr_quality = self.__doc_meta.get("alto_text_quality")
            if ocr_quality:
                self.__ocr_accuracy_schema = {
                    "comment": "Estimated OCR accuracy"
                }
                if self.__convert_ratio:
                    type_and_value = {
                        "@type": QUALITY_RATIO_TYPE,
                        "value": ocr_quality
                    }
                else:
                    type_and_value = {
                        "@type": "Text",
                        "value": ratio_to_percentage(ocr_quality)
                    }
                self.__ocr_accuracy_schema.update(type_and_value)
        return self.__ocr_accuracy_schema

    @property
    def text_quality_schema(self) -> dict:
        if not self.__text_quality_schema:
            text_quality = self.__doc_meta.get("text_quality")
            self.__text_quality_schema = {
                "comment": "Estimated n-gram-based text quality"
            }
            if self.__convert_ratio:
                type_and_value = {
                    "@type": QUALITY_RATIO_TYPE,
                    "value": text_quality
                }
            else:
                type_and_value = {
                    "@type": "Text",
                    "value": ratio_to_percentage(text_quality)
                }
            self.__text_quality_schema.update(type_and_value)
        return self.__text_quality_schema

    @property
    def language_schema(self) -> List[dict]:
        if not self.__language_schema:
            self.__language_schema = [
                {
                    "@type": "ISO 639-2",
                    "value": lang_to_iso639_2(
                        lang["language"],
                        unk_code=UNDEFINED_LANGUAGE_VALUE
                    )
                }
                for lang in self.__doc_meta["languages"]
                if lang["ratio"] >= self.__min_language_ratio
            ]
        return self.__language_schema

    @property
    def identifier_schema(self) -> List[dict]:
        if not self.__identifier_schema:
            identifiers = []
            if self.__sierra_id:
                identifiers.append(
                    {
                        "@type": "Identifier",
                        "qualifier": "OPAC",
                        "value": self.__sierra_id
                    }
                )
            if self.__permalink:
                identifiers.append(
                    {
                        "@type": "Identifier",
                        "qualifier": "Permalink",
                        "value": self.__permalink
                    }
                )
            if self.__generated_id:
                identifiers.append(
                    {
                        "@type": "Identifier",
                        "qualifier": self.__generated_id_type,
                        "value": self.__generated_id
                    }
                )
            self.__identifier_schema = identifiers

        return self.__identifier_schema

    @property
    def origin_schema(self) -> dict:
        if not self.__origin_schema:
            self.__origin_schema = {
                "@type": "Text",
                "value": self.origin,
                "comment": "Origin"
            }
        return self.__origin_schema


class DIGARSchemaConverter:
    def __init__(
            self,
            digitizer_output: dict,
            generated_id: str,
            sierra_id: str = "",
            permalink: str = "",
            generated_id_type: str = "CustomID",
            min_language_ratio: float = 0.2,
            convert_ratio: bool = False
    ) -> NoReturn:
        """ Initialize DIGARSchemaConverter object. 
        
        Parameters
        ----------
        digitizer_output: dict
            Raw output of rara-digitizer (https://pypi.org/project/rara-digitizer/).
        generated_id: str
            Some non-standard/generated document identifier used in ID fields.
        sierra_id: str
            Document's corresponding Sierra ID.
        permalink: str
            Permanent link, where the document can be accessed.
        generated_id_type: str
            Method / type of generated ID (e.g. 'UUID')
        min_language_ratio: float
            Cutoff ratio for languages. If ratio for some language
            does not exceed the set threshold, the language will not
            be added to the final output.
        convert_ratio: bool
            If enabled, all ratios are converted into percentages.
          
        """
        self.__digitizer_output: dict = digitizer_output
        self.__min_language_ratio: float = min_language_ratio
        self.__convert_ratio: bool = convert_ratio
        self.__sierra_id: str = sierra_id
        self.__generated_id: str = generated_id
        self.__permalink: str = permalink.removesuffix("/")
        self.__generated_id_type: str = generated_id_type
        self.__texts: List[dict] = []
        self.__images: List[dict] = []
        self.__doc_meta: dict = {}
        self.__page_mappings: List[dict] = []
        self.__dcterms_haspart: dict = {}
        self.__dcterms_conforms_to: dict = {}
        self.__dc_language: dict = {}
        self.__dc_origin: dict = {}
        self.__dc_identifier: List[dict] = []
        self.__doc_id: str = ""

        self.__doc_schemas = DocSchemas(
            doc_meta=self.doc_meta,
            sierra_id=self.__sierra_id,
            generated_id=self.__generated_id,
            permalink=self.__permalink,
            min_language_ratio=self.__min_language_ratio,
            convert_ratio=self.__convert_ratio,
            generated_id_type=self.__generated_id_type
        )
        self.__digar_schema: dict = {}

    def _get_page_number(self, page_content: dict) -> int:
        """ Retrieves page number from image or text object.
        """
        _segments = page_content["texts"] + page_content["images"]
        _first_segment = _segments[0]
        if "start_page" in _first_segment:
            page_number = _first_segment.get("start_page")
        elif "page" in _first_segment:
            page_number = _first_segment.get("page")
        return page_number

    @property
    def doc_id(self) -> str:
        """ Retrieves document ID to use for generating
        page and segment ids. Preference order:
        1. permalink; 2. sierra_id; 3. generated document id
        """
        if not self.__doc_id:
            if self.__permalink:
                self.__doc_id = self.__permalink
            elif self.__sierra_id:
                self.__doc_id = self.__sierra_id
            else:
                self.__doc_id = self.__generated_id
        return self.__doc_id

    @property
    def texts(self) -> List[dict]:
        if not self.__texts:
            self.__texts = self.__digitizer_output.get("texts")
        return self.__texts

    @property
    def images(self) -> List[dict]:
        if not self.__images:
            self.__images = self.__digitizer_output.get("images")
        return self.__images

    @property
    def doc_meta(self) -> dict:
        if not self.__doc_meta:
            self.__doc_meta = self.__digitizer_output.get("doc_meta")
        return self.__doc_meta

    @property
    def page_mappings(self) -> List[dict]:
        if not self.__page_mappings:
            mapped = defaultdict(lambda: defaultdict(list))
            for text in self.texts:
                mapped[text["start_page"]]["texts"].append(text)
            for img in self.images:
                mapped[img["page"]]["images"].append(img)

            self.__page_mappings = [
                v for k, v in sorted(list(mapped.items()), key=lambda x: x[0])
            ]
        return self.__page_mappings

    @property
    def dcterms_haspart(self) -> dict:
        if not self.__dcterms_haspart:
            self.__dcterms_haspart = {
                "dcterms:hasPart": [
                    PageSchema(
                        page_texts=page["texts"],
                        page_images=page["images"],
                        page_number=self._get_page_number(page),
                        doc_id=self.doc_id
                    ).schema
                    for page in self.page_mappings
                ]
            }
        return self.__dcterms_haspart

    @property
    def dcterms_conforms_to(self) -> dict:
        if not self.__dcterms_conforms_to:
            schema_content = [
                self.__doc_schemas.text_quality_schema,
            ]
            # Add OCR Accuracy only when it is not empty:
            if self.__doc_schemas.ocr_accuracy_schema:
                schema_content.append(self.__doc_schemas.ocr_accuracy_schema)
            self.__dcterms_conforms_to = {
                "dcterms:conformsTo": schema_content
            }
        return self.__dcterms_conforms_to

    @property
    def dc_language(self) -> dict:
        if not self.__dc_language:
            self.__dc_language = {
                "dc:language": self.__doc_schemas.language_schema
            }
        return self.__dc_language

    @property
    def dc_origin(self) -> dict:
        if not self.__dc_origin:
            self.__dc_origin = {
                "dcterms:provenance": self.__doc_schemas.origin_schema
            }
        return self.__dc_origin

    @property
    def dc_identifier(self) -> List[dict]:
        if not self.__dc_identifier:
            self.__dc_identifier = {
                "dc:identifier": self.__doc_schemas.identifier_schema
            }
        return self.__dc_identifier

    @property
    def digar_schema(self) -> dict:
        if not self.__digar_schema:
            self.__digar_schema = {}
            self.__digar_schema.update(self.dcterms_conforms_to)
            self.__digar_schema.update(self.dcterms_haspart)
            self.__digar_schema.update(self.dc_language)
            self.__digar_schema.update(self.dc_origin)
            self.__digar_schema.update(self.dc_identifier)
        return self.__digar_schema



import regex as re
import estnltk
import nltk
import logging
from rara_tools.parsers.tools.russian_transliterator import Transliterate
from rara_tools.constants.parsers import KeywordType, LOGGER
from typing import List, NoReturn
from abc import abstractmethod


nltk.download("punkt_tab")


class PersonalName:
    """ Wraps generating and accessing main name forms.
    """
    def __init__(self, name: str) -> NoReturn:
        """ Initializes PersonName object.

        Parameters
        -----------
        name: str
            Personal name. Expects one of the following formats:
            '<first name> <last name>' or '<last name>, <first_name>', e.g:
            'Uku Tamm' or 'Tamm, Uku'.
        """
        self.__original_name: str = name
        self.__name: dict = {}
        self.__last_comma_first: str = ""
        self.__first_last: str = ""

    @property
    def first_name(self) -> str:
        return self.name.get("first_name")

    @property
    def last_name(self) -> str:
        return self.name.get("last_name")

    @property
    def name(self) -> dict:
        if not self.__name:
            last_name = ""
            first_name = ""
            if "," in self.__original_name:
                try:
                    last_name, first_name = self.__original_name.split(",")
                except Exception as e:
                    LOGGER.error(
                        f"Parsing personal name {self.__original_name} " \
                        f"failed with error: {e}."
                    )
            else:
                name_tokens = [
                    t.strip()
                    for t in self.__original_name.split()
                    if t.strip()
                ]
                if len(name_tokens) > 1:
                    last_name = name_tokens[-1]
                    first_name = " ".join(name_tokens[:-1])
            self.__name = {
                "first_name": first_name.strip(),
                "last_name": last_name.strip()
            }
        return self.__name

    @property
    def last_comma_first(self) -> str:
        if not self.__last_comma_first:
            if self.last_name or self.first_name:
                self.__last_comma_first = f"{self.last_name}, {self.first_name}"
        return self.__last_comma_first.strip()

    @property
    def first_last(self) -> str:
        if not self.__first_last:
            self.__first_last = f"{self.first_name} {self.last_name}"
        return self.__first_last.strip()


class Normalizer:
    """ Class for handling general methods for string
    normalizations and variations generation.
    """
    def __init__(self, entity: str) -> NoReturn:
        """ Initializes Normalizer object.

        Parameters
        -----------
        entity: str
            Entity (keyword, person etc) to normalize.
        """
        self.__entity: str = entity
        self.__lemmatized_entity: str = ""
        self.__cleaned_entity: str = ""


    @staticmethod
    def has_cyrillic(entity: str) -> bool:
        return bool(re.search("[а-яА-Я]", entity))

    @staticmethod
    def transliterate(entity: str) -> str:
        transliterator = Transliterate()
        transliteration = transliterator([entity])[0]
        return transliteration

    @staticmethod
    def lemmatize(entity: str) -> str:
        layer = estnltk.Text(entity).tag_layer()
        lemma_list = [l[0] for l in list(layer.lemma)]
        lemmatized_entity = " ".join(lemma_list)
        return lemmatized_entity

    @staticmethod
    def remove_parenthesized_info(entity: str) -> str:
        clean_entity = re.sub(r"[(][^)]+[)]", "", entity)
        return clean_entity.strip()

    @staticmethod
    def clean_entity(entity: str) -> str:
        clean_entity = Normalizer.remove_parenthesized_info(entity)
        return clean_entity

    @property
    def lemmatized_entity(self) -> str:
        if not self.__lemmatized_entity:
            self.__lemmatized_entity = Normalizer.lemmatize(self.__entity)
        return self.__lemmatized_entity

    @property
    def cleaned_entity(self) -> str:
        if not self.__cleaned_entity:
            self.__cleaned_entity = Normalizer.clean_entity(self.__entity)
        return self.__cleaned_entity

    @abstractmethod
    def variations(self) -> List[str]:
        pass


class PersonNormalizer(Normalizer):
    """ Class for handling person-specific methods for string
    normalizations and variations generation.
    """
    def __init__(self, name: str) -> NoReturn:
        """ Initializes PersonNormalizer object.

        Parameters
        -----------
        name: str
            Personal name to normalize / generate variations for.
        """
        super().__init__(entity=name)
        self.__name: str = name
        self.__name_object: PersonalName = PersonalName(name)
        self.__variations: List[str] = []


    @property
    def variations(self) -> List[str]:
        if not self.__variations:
            LOGGER.debug(f"Generating variations for name {self.__name}.")
            variations = []
            variations.append(self.__name_object.last_comma_first)
            variations.append(self.__name_object.first_last)

            if Normalizer.has_cyrillic(self.__name):
                LOGGER.debug(
                    f"Detected cyrillic in the original name '{self.__name}'. " \
                    f"Generating a transliterated latin version."
                )
                transliterations = [
                    Normalizer.transliterate(name)
                    for name in variations
                ]
                variations.extend(transliterations)

            # Guarantee adding one-word names as well
            if self.__name not in variations:
                variations.append(self.__name)
            _variations = [v.strip() for v in variations if v.strip()]
            self.__variations = list(set(_variations))
            LOGGER.debug(
                f"Generated the following variations for name '{self.__name}': " \
                f"{self.__variations}."
            )
        return self.__variations



class KeywordNormalizer(Normalizer):
    """ Class for handling keyword-specific methods for string
    normalizations and variations generation.
    """
    def __init__(self, keyword: str, keyword_type: str = "") -> NoReturn:
        """ Initializes KeywordNormalizer object.

        Parameters
        -----------
        keyword: str
            keyword to normalize / generate variations for.
        keyword_type: str
            Keyword type. Should be one of the types specified in
            rara_tools.constants.parsers.KeywordType or "".

        """
        super().__init__(entity=keyword)
        self.__keyword: str = keyword
        self.__variations: List[str] = []
        self.__keyword_type: str = keyword_type
        self.__loc_substitutions_map: dict = {"v": "w", "V": "W"}

    def _transform_v_into_w(self, entity: str) -> str:
        for old_val, new_val in list(self.__loc_substitutions_map.items()):
            entity = re.sub(old_val, new_val, entity)
        return entity

    @property
    def loc_substitutions_as_str(self) -> str:
        subs = [
            f"'{old_val}' -> '{new_val}'"
            for old_val, new_val in list(self.__loc_substitutions_map.items())
        ]
        return ", ".join(subs)

    @property
    def variations(self) -> List[str]:
        if not self.__variations:
            LOGGER.debug(f"Generating variations for keyword {self.__keyword}.")
            variations = []
            variations.append(self.__keyword)
            variations.append(self.lemmatized_entity)
            variations.append(self.cleaned_entity)
            variations.append(Normalizer.lemmatize(self.cleaned_entity))
            # If keyword_type = LOC, add variations containing
            # v -> w replacements
            if self.__keyword_type == KeywordType.LOC:
                LOGGER.debug(
                    f"Detected keyword type = '{KeywordType.LOC}' -> " \
                    f"Adding variations with the following character " \
                    f"replacements: {self.loc_substitutions_as_str}."
                )
                v_w_transformations = [
                    self._transform_v_into_w(entity)
                    for entity in variations
                ]
                variations.extend(v_w_transformations)
            variations = list(set(variations))
            self.__variations = variations
            LOGGER.debug(
                f"Generated the following variations for keyword '{self.__keyword}': " \
                f"{self.__variations}."
            )
        return self.__variations



COMPONENT_KEY = "digitizer"


class ModelTypes:
    IMAGE_PROCESSOR = "image_processor"


class StatusKeys:
    DOWNLOAD_MODELS = "digitizer_download_models"
    CLEAN_UP = "digitizer_clean_up"
    ELASTICSEARCH_UPLOAD = "digitizer_elasticsearch_upload"
    UPLOAD = "s3_upload"
    DOWNLOAD = "digitizer_s3_download"
    OCR = "digitizer_ocr"


class Queue:
    IO = "io"
    DOWNLOAD = "download"
    FINISH = "finish"
    OCR = "ocr"


class Tasks:
    MODEL_UPDATE = "component_model_update"



class S3InputException(Exception):
    """Raised S3 input error."""

class S3InitException(Exception):
    """Raised S3 Error."""

class S3ConnectionException(Exception):
    """Raised S3 Bucket/Connection Error."""

class S3DownloadException(Exception):
    """Raised S3 Download Error."""


class ElasticsearchException(Exception):
    """Raised Elasticsearch Error."""

class TaskReporterException(Exception):
    """Raised TaskReporter Error."""

class SierraResponseConverterException(Exception):
    """Raised SierraResponseConverter Error."""
